use itertools::Itertools;

use crate::util::Dir;

#[derive(Debug)]
pub enum TurnDir {
    CW,
    CCW,
}

#[derive(Debug)]
pub enum PassInstr {
    Move(usize),
    Turn(TurnDir),
}

#[derive(Debug, Clone)]
pub struct Pose {
    pub pos: (usize, usize),
    pub orientation: Dir,
}

impl Pose {
    pub fn new(grid: &SparseGrid) -> Pose {
        let col = grid.grid[0]
            .iter()
            .position(|t| if let Some(occ) = t { !occ } else { false })
            .unwrap();
        Pose {
            pos: (0, col),
            orientation: Dir::Right,
        }
    }
}

pub enum MoveResult {
    Blocked,
    BlockedByWrapping,
    WrappedAround,
    Success,
}

impl MoveResult {
    pub fn was_blocked(&self) -> bool {
        matches!(self, MoveResult::Blocked | MoveResult::BlockedByWrapping)
    }

    pub fn wrapped_around(&self) -> bool {
        matches!(
            self,
            MoveResult::WrappedAround | MoveResult::BlockedByWrapping
        )
    }
}

pub trait Navigable {
    type Pose;
    type NavigationInstruction;

    fn forward(&self, from: &Self::Pose, steps: usize) -> (Self::Pose, MoveResult);
    fn exec_instruction(
        &self,
        from: &Self::Pose,
        inst: &Self::NavigationInstruction,
    ) -> (Self::Pose, MoveResult);
}

#[derive(Debug)]
pub struct SparseGrid {
    pub grid: Vec<Vec<Option<bool>>>,
    pub height: usize,
    pub width: usize,
    pub row_bounds: Vec<(usize, usize)>,
    pub col_bounds: Vec<(usize, usize)>,
}

impl Navigable for SparseGrid {
    type Pose = Pose;
    type NavigationInstruction = PassInstr;

    fn forward(&self, from: &Pose, steps: usize) -> (Pose, MoveResult) {
        let mut pose = from.clone();
        let pos = &mut pose.pos;
        let mut status = MoveResult::Success;
        for _ in 0..steps {
            let next_pos = match from.orientation {
                Dir::Left => {
                    let bounds = &self.row_bounds[pos.0];
                    let next_col = if pos.1 == 0 || pos.1 == bounds.0 {
                        status = MoveResult::WrappedAround;
                        bounds.1
                    } else {
                        pos.1 - 1
                    };
                    (pos.0, next_col)
                }
                Dir::Up => {
                    let bounds = &self.col_bounds[pos.1];
                    let next_row = if pos.0 == 0 || pos.0 == bounds.0 {
                        status = MoveResult::WrappedAround;
                        bounds.1
                    } else {
                        pos.0 - 1
                    };
                    (next_row, pos.1)
                }
                Dir::Right => {
                    let bounds = &self.row_bounds[pos.0];
                    let next_col = if pos.1 == bounds.1 {
                        status = MoveResult::WrappedAround;
                        bounds.0
                    } else {
                        pos.1 + 1
                    };
                    (pos.0, next_col)
                }
                Dir::Down => {
                    let bounds = &self.col_bounds[pos.1];
                    let next_row = if pos.0 == bounds.1 {
                        status = MoveResult::WrappedAround;
                        bounds.0
                    } else {
                        pos.0 + 1
                    };
                    (next_row, pos.1)
                }
            };

            if self.grid[next_pos.0][next_pos.1].unwrap() {
                // We're about to hit a wall
                if status.wrapped_around() {
                    return (pose, MoveResult::BlockedByWrapping);
                } else {
                    return (pose, MoveResult::Blocked);
                }
            }
            *pos = next_pos;
        }

        (pose, status)
    }

    fn exec_instruction(
        &self,
        from: &Self::Pose,
        inst: &Self::NavigationInstruction,
    ) -> (Self::Pose, MoveResult) {
        match inst {
            PassInstr::Move(n) => self.forward(from, *n),
            PassInstr::Turn(dir) => (
                Pose {
                    pos: from.pos,
                    orientation: match dir {
                        TurnDir::CW => from.orientation.cw(),
                        TurnDir::CCW => from.orientation.ccw(),
                    },
                },
                MoveResult::Success,
            ),
        }
    }
}

pub fn parse_map_and_path(input: &str) -> (SparseGrid, Vec<PassInstr>) {
    let lines = input.lines().collect_vec();

    let mut grid: Vec<Vec<Option<bool>>> = Vec::new();
    let mut width = 0;
    for line in &lines[..lines.len() - 2] {
        width = width.max(line.len());
        let grid_line = line
            .chars()
            .map(|c| match c {
                ' ' => None,
                '.' => Some(false),
                '#' => Some(true),
                _ => panic!("Unknown grid character {}", c),
            })
            .collect();
        grid.push(grid_line);
    }

    let height = grid.len();
    let mut row_bounds = vec![(0, 0); height];
    for row in 0..height {
        let row_width = grid[row].len();
        if row_width < width {
            grid[row].extend((row_width..width).map(|_| None));
        }
        let lower = grid[row].iter().position(Option::is_some).unwrap();
        let upper = width - 1 - grid[row].iter().rev().position(Option::is_some).unwrap();
        row_bounds[row] = (lower, upper);
    }

    let mut col_bounds = vec![(0, 0); width];
    for col in 0..width {
        let col_it = (0..height).map(|row| grid[row][col]);
        let lower = col_it.clone().position(|t| t.is_some()).unwrap();
        let upper = height - 1 - col_it.rev().position(|t| t.is_some()).unwrap();
        col_bounds[col] = (lower, upper);
    }

    let grid = SparseGrid {
        grid,
        height,
        width,
        row_bounds,
        col_bounds,
    };

    let mut instructions = Vec::new();
    let mut inst_begin = 0;
    let inst_line = lines[lines.len() - 1].as_bytes();
    while inst_begin < inst_line.len() {
        let mut i = inst_begin;
        let c = inst_line[i];
        if c.is_ascii_digit() {
            while i < inst_line.len() && inst_line[i].is_ascii_digit() {
                i += 1;
            }
            let dist = std::str::from_utf8(&inst_line[inst_begin..i])
                .unwrap()
                .parse()
                .unwrap();
            instructions.push(PassInstr::Move(dist));
            inst_begin = i;
            continue;
        }

        let dir = match c {
            b'L' => TurnDir::CCW,
            b'R' => TurnDir::CW,
            _ => panic!("Unknown turn direction {}", c),
        };
        instructions.push(PassInstr::Turn(dir));
        inst_begin += 1;
    }

    (grid, instructions)
}
